#!/usr/bin/env python3
"""
Memory Guardian
Real-time memory analysis and protection system.
Parameters:
    - pid: Process ID to monitor (optional)
    - mode: Protection mode (monitor, protect, forensic)
Requires: admin/root privileges
"""
import psutil
import logging
from datetime import datetime
import json
from pathlib import Path
import threading
import queue
import time
import re
import platform
import ctypes
import struct
import volatility3.framework.interfaces.configuration as interfaces
import volatility3.framework.contexts as contexts
import volatility3.framework.automagic as automagic
import yara
import argparse

class MemoryGuardian:
    def __init__(self, pid=None, mode='monitor'):
        self.pid = pid
        self.mode = mode
        self.is_windows = platform.system().lower() == 'windows'
        
        self.event_queue = queue.Queue()
        self.alert_queue = queue.Queue()
        self.stop_event = threading.Event()
        
        # Memory patterns to detect
        self.patterns = {
            'shellcode': [
                rb'\x55\x8B\xEC',  # Push EBP; MOV EBP, ESP
                rb'\x33\xC0',      # XOR EAX, EAX
                rb'\x90{10,}'      # NOP sled
            ],
            'credentials': [
                rb'password[=:]\s*\w+',
                rb'admin[=:]\s*\w+',
                rb'root[=:]\s*\w+'
            ],
            'network': [
                rb'(http|ftp|smtp)://',
                rb'\b(?:[0-9]{1,3}\.){3}[0-9]{1,3}\b'
            ]
        }
        
        self.setup_logging()
        self.initialize_yara_rules()
        
    def setup_logging(self):
        """Set up logging configuration"""
        log_dir = Path("memory_logs")
        log_dir.mkdir(exist_ok=True)
        
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s',
            handlers=[
                logging.FileHandler(log_dir / "memory.log"),
                logging.StreamHandler()
            ]
        )
        
    def initialize_yara_rules(self):
        """Initialize YARA rules for memory scanning"""
        rules = """
rule Shellcode {
    strings:
        $s1 = { 55 8B EC }  // Push EBP; MOV EBP, ESP
        $s2 = { 33 C0 }     // XOR EAX, EAX
        $s3 = { 90 90 90 90 90 } // NOP sled
    condition:
        any of them
}

rule ProcessInjection {
    strings:
        $api1 = "VirtualAllocEx" ascii wide
        $api2 = "WriteProcessMemory" ascii wide
        $api3 = "CreateRemoteThread" ascii wide
    condition:
        2 of them
}

rule CredentialHarvesting {
    strings:
        $cred1 = /password[=:]\s*\w+/ ascii wide
        $cred2 = /username[=:]\s*\w+/ ascii wide
        $cred3 = /login[=:]\s*\w+/ ascii wide
    condition:
        any of them
}
"""
        # Save rules to file
        rules_dir = Path("yara_rules")
        rules_dir.mkdir(exist_ok=True)
        rules_file = rules_dir / "memory_rules.yar"
        
        with open(rules_file, 'w') as f:
            f.write(rules)
            
        self.yara_rules = yara.compile(filepath=str(rules_file))
        
    def scan_process_memory(self, process):
        """Scan process memory regions"""
        try:
            # Get memory maps
            for mmap in process.memory_maps():
                try:
                    # Read memory region
                    memory_region = process.memory_region(
                        mmap.addr,
                        mmap.size
                    )
                    
                    # Scan with YARA rules
                    matches = self.yara_rules.match(data=memory_region)
                    if matches:
                        self.alert_queue.put({
                            'type': 'yara_match',
                            'pid': process.pid,
                            'process': process.name(),
                            'address': mmap.addr,
                            'rules': [m.rule for m in matches]
                        })
                        
                    # Pattern matching
                    for category, patterns in self.patterns.items():
                        for pattern in patterns:
                            if re.search(pattern, memory_region):
                                self.alert_queue.put({
                                    'type': 'pattern_match',
                                    'pid': process.pid,
                                    'process': process.name(),
                                    'address': mmap.addr,
                                    'category': category,
                                    'pattern': pattern
                                })
                                
                except Exception as e:
                    logging.debug(f"Error reading memory region: {str(e)}")
                    
        except Exception as e:
            logging.error(f"Error scanning process {process.pid}: {str(e)}")
            
    def protect_process_memory(self, process):
        """Apply memory protection measures"""
        try:
            if self.is_windows:
                # Use Windows API for memory protection
                kernel32 = ctypes.windll.kernel32
                
                # Get process handle
                PROCESS_ALL_ACCESS = 0x1F0FFF
                handle = kernel32.OpenProcess(
                    PROCESS_ALL_ACCESS,
                    False,
                    process.pid
                )
                
                if handle:
                    try:
                        # Protect memory regions
                        for mmap in process.memory_maps():
                            PAGE_READONLY = 0x02
                            kernel32.VirtualProtectEx(
                                handle,
                                mmap.addr,
                                mmap.size,
                                PAGE_READONLY,
                                ctypes.byref(ctypes.c_ulong())
                            )
                    finally:
                        kernel32.CloseHandle(handle)
                        
            else:
                # Use ptrace for Linux memory protection
                import ptrace.debugger
                debugger = ptrace.debugger.PtraceDebugger()
                process = debugger.addProcess(process.pid, False)
                process.cont()
                
        except Exception as e:
            logging.error(f"Error protecting process {process.pid}: {str(e)}")
            
    def dump_process_memory(self, process):
        """Create memory dump for forensic analysis"""
        try:
            dump_dir = Path("memory_dumps")
            dump_dir.mkdir(exist_ok=True)
            
            dump_file = dump_dir / f"dump_{process.pid}_{datetime.now().strftime('%Y%m%d_%H%M%S')}.dmp"
            
            if self.is_windows:
                # Use Windows API for memory dump
                kernel32 = ctypes.windll.kernel32
                
                PROCESS_ALL_ACCESS = 0x1F0FFF
                handle = kernel32.OpenProcess(
                    PROCESS_ALL_ACCESS,
                    False,
                    process.pid
                )
                
                if handle:
                    try:
                        # Create dump file
                        kernel32.MiniDumpWriteDump(
                            handle,
                            process.pid,
                            dump_file,
                            2,  # MiniDumpWithFullMemory
                            None,
                            None,
                            None
                        )
                    finally:
                        kernel32.CloseHandle(handle)
                        
            else:
                # Use GDB for Linux memory dump
                subprocess.run([
                    'gdb',
                    '--batch',
                    '-p', str(process.pid),
                    '-ex', f'dump memory {dump_file} 0 0xffffffffffffffff'
                ], check=True)
                
            return dump_file
            
        except Exception as e:
            logging.error(f"Error dumping process {process.pid}: {str(e)}")
            return None
            
    def analyze_memory_dump(self, dump_file):
        """Analyze memory dump using Volatility"""
        try:
            # Initialize Volatility
            context = contexts.Context()
            config = dict()
            
            # Add dump file to config
            config['automagic.LayerStacker.single_location'] = str(dump_file)
            
            # Run Volatility plugins
            plugins = [
                'windows.pslist.PsList',
                'windows.handles.Handles',
                'windows.modules.Modules'
            ]
            
            results = {}
            for plugin in plugins:
                try:
                    plugin_config = interfaces.configuration.path_join(
                        'plugins',
                        plugin
                    )
                    
                    automagics = automagic.available(context)
                    constructed = plugins.construct_plugin(
                        context,
                        automagics,
                        plugin,
                        config,
                        progress_callback=None
                    )
                    
                    results[plugin] = list(constructed.run())
                except Exception as e:
                    logging.error(f"Error running Volatility plugin {plugin}: {str(e)}")
                    
            return results
            
        except Exception as e:
            logging.error(f"Error analyzing memory dump: {str(e)}")
            return None
            
    def monitor_processes(self):
        """Monitor processes for memory changes"""
        while not self.stop_event.is_set():
            try:
                processes = []
                if self.pid:
                    # Monitor specific process
                    try:
                        processes = [psutil.Process(self.pid)]
                    except psutil.NoSuchProcess:
                        logging.error(f"Process {self.pid} not found")
                        break
                else:
                    # Monitor all processes
                    processes = psutil.process_iter(['pid', 'name'])
                    
                for process in processes:
                    if self.stop_event.is_set():
                        break
                        
                    try:
                        # Scan process memory
                        self.scan_process_memory(process)
                        
                        # Apply protection in protect mode
                        if self.mode == 'protect':
                            self.protect_process_memory(process)
                            
                        # Create memory dump in forensic mode
                        elif self.mode == 'forensic':
                            dump_file = self.dump_process_memory(process)
                            if dump_file:
                                analysis = self.analyze_memory_dump(dump_file)
                                if analysis:
                                    self.event_queue.put({
                                        'type': 'memory_analysis',
                                        'pid': process.pid,
                                        'process': process.name(),
                                        'dump_file': str(dump_file),
                                        'analysis': analysis
                                    })
                                    
                    except (psutil.NoSuchProcess, psutil.AccessDenied):
                        continue
                        
                time.sleep(1)  # Prevent high CPU usage
                
            except Exception as e:
                logging.error(f"Error monitoring processes: {str(e)}")
                
    def handle_alerts(self):
        """Handle memory analysis alerts"""
        while not self.stop_event.is_set():
            try:
                alert = self.alert_queue.get(timeout=1)
                logging.warning(
                    f"Memory Alert - Type: {alert['type']}, "
                    f"Process: {alert['process']} (PID: {alert['pid']})"
                )
                
                # Take action based on alert type
                if alert['type'] == 'yara_match':
                    if self.mode in ['protect', 'forensic']:
                        process = psutil.Process(alert['pid'])
                        dump_file = self.dump_process_memory(process)
                        if dump_file:
                            alert['dump_file'] = str(dump_file)
                            
                elif alert['type'] == 'pattern_match':
                    if alert['category'] == 'shellcode':
                        if self.mode == 'protect':
                            process = psutil.Process(alert['pid'])
                            self.protect_process_memory(process)
                            
            except queue.Empty:
                continue
                
    def start_monitoring(self):
        """Start memory monitoring"""
        logging.info("Starting Memory Guardian")
        logging.info(f"Mode: {self.mode}")
        if self.pid:
            logging.info(f"Monitoring PID: {self.pid}")
            
        # Start monitoring thread
        monitor_thread = threading.Thread(target=self.monitor_processes)
        monitor_thread.start()
        
        # Start alert handler
        alert_thread = threading.Thread(target=self.handle_alerts)
        alert_thread.start()
        
        try:
            while True:
                # Print status
                alert_count = sum(1 for _ in self.alert_queue.queue)
                print(f"\rActive alerts: {alert_count}", end="")
                time.sleep(1)
                
        except KeyboardInterrupt:
            logging.info("\nStopping Memory Guardian...")
            self.stop_event.set()
            
            # Wait for threads to finish
            monitor_thread.join()
            alert_thread.join()
            
            # Save final state
            self.save_state()
            
    def save_state(self):
        """Save monitoring state and results"""
        state = {
            'timestamp': datetime.now().isoformat(),
            'mode': self.mode,
            'pid': self.pid
        }
        
        # Get all events
        events = []
        while not self.event_queue.empty():
            events.append(self.event_queue.get())
        state['events'] = events
        
        # Get all alerts
        alerts = []
        while not self.alert_queue.empty():
            alerts.append(self.alert_queue.get())
        state['alerts'] = alerts
        
        # Save to file
        output_dir = Path("memory_logs")
        output_file = output_dir / f"state_{datetime.now().strftime('%Y%m%d_%H%M%S')}.json"
        with open(output_file, 'w') as f:
            json.dump(state, f, indent=2)
            
        logging.info(f"State saved to {output_file}")
        
        # Print summary
        print("\nMemory Guardian Summary")
        print("=" * 50)
        print(f"Mode: {self.mode}")
        if self.pid:
            print(f"Monitored PID: {self.pid}")
        print(f"Total Alerts: {len(alerts)}")
        print(f"Total Events: {len(events)}")
        
        if alerts:
            print("\nAlert Summary:")
            alert_types = {}
            for alert in alerts:
                alert_types[alert['type']] = alert_types.get(alert['type'], 0) + 1
            for alert_type, count in alert_types.items():
                print(f"{alert_type}: {count}")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Memory Guardian")
    parser.add_argument("--pid", type=int, help="Process ID to monitor")
    parser.add_argument("--mode", choices=['monitor', 'protect', 'forensic'],
                      default='monitor', help="Protection mode")
    args = parser.parse_args()
    
    guardian = MemoryGuardian(args.pid, args.mode)
    guardian.start_monitoring()
